@@ -7,7 +7,7 @@ use Carp ();
 use IO::Handle ();
 use Exporter ();
 
-our $VERSION = '0.016';
+our $VERSION = '0.017';
 
 use constant DEFAULT_REQUEST_BODY_LIMIT => 16777216;
 use constant DEFAULT_REQUEST_BODY_BUFFER => 262144;
@@ -95,7 +95,10 @@ my %HTTP_STATUS = (
   sub cgi (&) {
     my ($handler) = @_;
     $cgi ||= bless {pid => $$}, __PACKAGE__;
-    _debug_command($cgi, [@ARGV]) if @ARGV and !defined $ENV{REQUEST_METHOD};
+    if (@ARGV and !defined $ENV{REQUEST_METHOD}) {
+      require CGI::Tiny::_Debug;
+      CGI::Tiny::_Debug::debug_command($cgi, [@ARGV]);
+    }
     my ($error, $errored);
     {
       local $@;
@@ -146,6 +149,7 @@ sub _handle_error {
 sub set_error_handler          { $_[0]{on_error} = $_[1]; $_[0] }
 sub set_request_body_buffer    { $_[0]{request_body_buffer} = $_[1]; $_[0] }
 sub set_request_body_limit     { $_[0]{request_body_limit} = $_[1]; $_[0] }
+sub set_discard_form_files     { $_[0]{discard_form_files} = @_ < 2 ? 1 : $_[1]; $_[0] }
 sub set_multipart_form_charset { $_[0]{multipart_form_charset} = $_[1]; $_[0] }
 sub set_input_handle           { $_[0]{input_handle} = $_[1]; $_[0] }
 sub set_output_handle          { $_[0]{output_handle} = $_[1]; $_[0] }
@@ -370,9 +374,8 @@ sub _body_multipart {
   my ($self) = @_;
   unless (exists $self->{body_parts}) {
     $self->{body_parts} = [];
-    my ($boundary_quoted, $boundary_unquoted) = $ENV{CONTENT_TYPE} =~ m/;\s*boundary\s*=\s*(?:"((?:\\[\\"]|[^"])+)"|([^";]+))/i;
-    $boundary_quoted =~ s/\\([\\"])/$1/g if defined $boundary_quoted;
-    my $boundary = defined $boundary_quoted ? $boundary_quoted : $boundary_unquoted;
+    require CGI::Tiny::Multipart;
+    my $boundary = CGI::Tiny::Multipart::extract_multipart_boundary($ENV{CONTENT_TYPE});
     unless (defined $boundary) {
       $self->{response_status} = "400 $HTTP_STATUS{400}" unless $self->{headers_rendered};
       die "Malformed multipart/form-data request\n";
@@ -385,10 +388,12 @@ sub _body_multipart {
     } else {
       $length = $self->_body_length;
       $input = defined $self->{input_handle} ? $self->{input_handle} : *STDIN;
-      binmode $input;
     }
 
-    my $parts = _parse_multipart($input, $length, $boundary, $self->{request_body_buffer} || $ENV{CGI_TINY_REQUEST_BODY_BUFFER});
+    my $parts = CGI::Tiny::Multipart::parse_multipart_form_data($input, $length, $boundary, {
+      buffer_size => $self->{request_body_buffer} || $ENV{CGI_TINY_REQUEST_BODY_BUFFER},
+      discard_files => $self->{discard_form_files} || $ENV{CGI_TINY_DISCARD_FORM_FILES},
+    });
     unless (defined $parts) {
       $self->{response_status} = "400 $HTTP_STATUS{400}" unless $self->{headers_rendered};
       die "Malformed multipart/form-data request\n";
@@ -626,7 +631,7 @@ sub response_status_code {
 sub _response_headers {
   my ($self, $type, $content_length, $location) = @_;
   my $headers_str = '';
-  return $headers_str if defined $ENV{CGI_TINY_DEBUG_METHOD} and !$ENV{CGI_TINY_DEBUG_VERBOSE};
+  return $headers_str if defined $self->{debug_method} and !$self->{debug_verbose};
   my %headers_set;
   foreach my $header (@{$self->{response_headers} || []}) {
     my ($name, $value) = @$header;
@@ -700,131 +705,6 @@ sub _json {
   return $self->{json};
 }
 
-sub _parse_multipart {
-  my ($input, $length, $boundary, $buffer_size) = @_;
-  $buffer_size = 0 + ($buffer_size || DEFAULT_REQUEST_BODY_BUFFER);
-  my $buffer = "\r\n";
-  my $next_boundary = "\r\n--$boundary\r\n";
-  my $end_boundary = "\r\n--$boundary--";
-  my (%state, @parts);
-  READER: while ($length > 0) {
-    if (ref $input eq 'SCALAR') {
-      $buffer .= $$input;
-      $length = 0;
-    } else {
-      my $chunk = $length < $buffer_size ? $length : $buffer_size;
-      last unless my $read = read $input, $buffer, $chunk, length $buffer;
-      $length -= $read;
-    }
-
-    unless ($state{parsing_headers} or $state{parsing_body}) {
-      my $next_pos = index $buffer, $next_boundary;
-      my $end_pos = index $buffer, $end_boundary;
-      if ($next_pos >= 0 and ($end_pos < 0 or $end_pos > $next_pos)) {
-        substr $buffer, 0, $next_pos + length($next_boundary), '';
-        $state{parsing_headers} = 1;
-        push @parts, $state{part} = {headers => {}, name => undef, filename => undef, size => 0};
-      } elsif ($end_pos >= 0) {
-        $state{done} = 1;
-        last; # end of multipart data
-      } else {
-        next; # read more to find start of multipart data
-      }
-    }
-
-    while (length $buffer) {
-      if ($state{parsing_headers}) {
-        while ((my $pos = index $buffer, "\r\n") >= 0) {
-          if ($pos == 0) { # end of headers
-            $state{parsing_headers} = 0;
-            $state{parsing_body} = 1;
-            $state{parsed_optional_crlf} = 0;
-            last;
-          }
-
-          my $header = substr $buffer, 0, $pos + 2, '';
-          my ($name, $value) = split /\s*:\s*/, $header, 2;
-          return undef unless defined $value;
-          $value =~ s/\s*\z//;
-
-          $state{part}{headers}{lc $name} = $value;
-          if (lc $name eq 'content-disposition') {
-            while ($value =~ m/;\s*([^=\s]+)\s*=\s*(?:"((?:\\[\\"]|[^"])*)"|([^";]*))/ig) {
-              my ($field_name, $field_quoted, $field_unquoted) = ($1, $2, $3);
-              next unless lc $field_name eq 'name' or lc $field_name eq 'filename';
-              $field_quoted =~ s/\\([\\"])/$1/g if defined $field_quoted;
-              $state{part}{lc $field_name} = defined $field_quoted ? $field_quoted : $field_unquoted;
-            }
-          }
-        }
-        next READER if $state{parsing_headers}; # read more to find end of headers
-      } else {
-        my $append = '';
-        my $next_pos = index $buffer, $next_boundary;
-        my $end_pos = index $buffer, $end_boundary;
-        if ($next_pos >= 0 and ($end_pos < 0 or $end_pos > $next_pos)) {
-          if (!$state{parsed_optional_crlf} and $next_pos >= 2) {
-            substr $buffer, 0, 2, '';
-            $next_pos -= 2;
-            $state{parsed_optional_crlf} = 1;
-          }
-          $append = substr $buffer, 0, $next_pos, '';
-          substr $buffer, 0, length($next_boundary), '';
-          $state{parsing_body} = 0;
-          $state{parsing_headers} = 1;
-        } elsif ($end_pos >= 0) {
-          if (!$state{parsed_optional_crlf} and $end_pos >= 2) {
-            substr $buffer, 0, 2, '';
-            $end_pos -= 2;
-            $state{parsed_optional_crlf} = 1;
-          }
-          $append = substr $buffer, 0, $end_pos; # no replacement, we're done here
-          $state{parsing_body} = 0;
-          $state{done} = 1;
-        } elsif (length($buffer) > length($next_boundary) + 2) {
-          if (!$state{parsed_optional_crlf}) {
-            substr $buffer, 0, 2, '';
-            $state{parsed_optional_crlf} = 1;
-          }
-          $append = substr $buffer, 0, length($buffer) - length($next_boundary), '';
-        }
-
-        if (defined $state{part}{filename}) {
-          # create temp file even if empty
-          unless (defined $state{part}{file}) {
-            require File::Temp;
-            $state{part}{file} = File::Temp->new;
-            binmode $state{part}{file};
-          }
-          if (length $append) {
-            $state{part}{file}->print($append);
-            $state{part}{size} += length $append;
-          }
-          unless ($state{parsing_body}) { # finalize temp file
-            $state{part}{file}->flush;
-            seek $state{part}{file}, 0, 0;
-          }
-        } else {
-          $state{part}{content} = '' unless defined $state{part}{content};
-          if (length $append) {
-            $state{part}{content} .= $append;
-            $state{part}{size} += length $append;
-          }
-        }
-
-        last READER if $state{done};         # end of multipart data
-        next READER if $state{parsing_body}; # read more to find end of part
-
-        # new part started
-        push @parts, $state{part} = {headers => {}, name => undef, filename => undef, size => 0};
-      }
-    }
-  }
-  return undef unless $state{done};
-
-  return \@parts;
-}
-
 {
   my @DAYS_OF_WEEK = qw(Sun Mon Tue Wed Thu Fri Sat);
   my @MONTH_NAMES = qw(Jan Feb Mar Apr May Jun Jul Aug Sep Oct Nov Dec);
@@ -868,72 +748,4 @@ sub _parse_multipart {
   sub escape_html { (my $escaped = $_[0]) =~ s/([&<>"'])/$ESCAPES{$1}/ge; $escaped }
 }
 
-{
-  my %methods = (get => 1, head => 1, post => 1, put => 1, delete => 1);
-  sub _debug_command {
-    my ($cgi, $argv) = @_;
-    return 1 unless @$argv;
-    my $command = shift @$argv;
-    if (exists $methods{$command}) {
-      $ENV{CGI_TINY_DEBUG_METHOD} = uc $command;
-      $ENV{CGI_TINY_DEBUG_VERBOSE} = 1 if uc($command) eq 'HEAD';
-
-      require Getopt::Long;
-      Getopt::Long::Configure('default', 'gnu_getopt', 'no_ignore_case');
-      Getopt::Long::GetOptionsFromArray($argv,
-        'content|c=s' => \my $content,
-        'cookie|C=s' => \my @cookies,
-        'header|H=s' => \my @headers,
-        'verbose|v' => \$ENV{CGI_TINY_DEBUG_VERBOSE},
-      ) or die "Failed to parse debug command options\n";
-
-      my ($path) = @$argv;
-
-      if (defined $content) {
-        open my $in_fh, '<', \$content or die "Failed to open in-memory handle to request content: $!\n";
-        $cgi->set_input_handle($in_fh);
-      }
-
-      foreach my $header (@headers) {
-        my ($name, $value) = split /\s*:\s*/, $header, 2;
-        next unless defined $value;
-        $name =~ tr/-/_/;
-        if (defined $ENV{"HTTP_\U$name"}) {
-          $ENV{"HTTP_\U$name"} .= ", $value";
-        } else {
-          $ENV{"HTTP_\U$name"} = $value;
-        }
-      }
-
-      $ENV{HTTP_CONTENT_LENGTH} = length $content if defined $content;
-      $ENV{HTTP_CONTENT_LENGTH} = '' unless defined $ENV{HTTP_CONTENT_LENGTH};
-      $ENV{HTTP_CONTENT_TYPE} = 'application/octet-stream' if $ENV{HTTP_CONTENT_LENGTH} and !defined $ENV{HTTP_CONTENT_TYPE};
-      $ENV{HTTP_COOKIE} = join '; ', @cookies if @cookies;
-
-      my $query;
-      ($path, $query) = split /\?/, $path, 2 if defined $path;
-      $ENV{AUTH_TYPE} = '' unless defined $ENV{AUTH_TYPE};
-      $ENV{CONTENT_LENGTH} = $ENV{HTTP_CONTENT_LENGTH};
-      $ENV{CONTENT_TYPE} = $ENV{HTTP_CONTENT_TYPE};
-      $ENV{GATEWAY_INTERFACE} = 'CGI/1.1' unless defined $ENV{GATEWAY_INTERFACE};
-      $ENV{PATH_INFO} = defined $path ? $path : '';
-      $ENV{PATH_TRANSLATED} = '' unless defined $ENV{PATH_TRANSLATED};
-      $ENV{QUERY_STRING} = defined $query ? $query : '';
-      $ENV{REMOTE_ADDR} = '127.0.0.1' unless defined $ENV{REMOTE_ADDR};
-      $ENV{REMOTE_HOST} = do { require Sys::Hostname; Sys::Hostname::hostname() } unless defined $ENV{REMOTE_HOST};
-      $ENV{REMOTE_IDENT} = '' unless defined $ENV{REMOTE_IDENT};
-      $ENV{REMOTE_USER} = '' unless defined $ENV{REMOTE_USER};
-      $ENV{REQUEST_METHOD} = uc $command;
-      $ENV{SCRIPT_NAME} = do { require File::Basename; '/' . File::Basename::basename($0) } unless defined $ENV{SCRIPT_NAME};
-      $ENV{SERVER_NAME} = do { require Sys::Hostname; Sys::Hostname::hostname() } unless defined $ENV{SERVER_NAME};
-      $ENV{SERVER_PORT} = 80 unless defined $ENV{SERVER_PORT};
-      $ENV{SERVER_PROTOCOL} = 'HTTP/1.0' unless defined $ENV{SERVER_PROTOCOL};
-      $ENV{SERVER_SOFTWARE} = "CGI::Tiny/$VERSION" unless defined $ENV{SERVER_SOFTWARE};
-    } else {
-      die "Unknown debug command $command\n";
-    }
-    return 1;
-  }
-}
-
 1;
